/*

		Written by Jeff Mersberger (jeff@mersberger.com)
        Access Log - tracks authentications

License: This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later version.

*/

#include "tac_plus.h"
#include <stdio.h>
#include <time.h>

static char *pUnknown = "unknown";

/* triggered by specifing access file = <filespec> in config */

/*time NAS_name NAC_address user type status*/

int log_access(AUTHEN_DATA *data)
{
	char	*filename;
	int		logfd = 0;
	char 	szOut[256], *pFrom, *pOutPos;
	int		cntOut, len;
	time_t t = time(NULL);
	char *ct = (char *) ctime(&t);

	if (session.accesslog_file) {
		filename = session.accesslog_file;
	} else {
#ifdef USE_DB
		if (session.db_default)
			return db_access(data);
#endif
		/* no need to go any further if file not specified */
		return 0;
	}

	if (!data) {
		return 0;
	}
	if (!data->NAS_id) {
		return 0;
	}

	if (ct) {
		ct[24] = '\0';
	} else {
		ct = "\t";
	}

	bzero(szOut,sizeof(szOut));
	pOutPos = szOut;
	cntOut = sizeof(szOut) - 1;			/* save space for terminating null */

	pFrom = ct;
	if (pFrom) {
		while (*pFrom && cntOut) {
			*(pOutPos++) = *(pFrom++);
			cntOut--;
		}
	}
	if (cntOut) { *(pOutPos++) = '\t'; cntOut--; }

	pFrom = data->NAS_id->NAS_name;
	if (pFrom == NULL) pFrom=pUnknown;
	while (*pFrom && cntOut) {
		*(pOutPos++) = *(pFrom++);
		cntOut--;
	}
	if (cntOut) { *(pOutPos++) = '\t'; cntOut--; }
    pFrom = data->NAS_id->NAS_port;
	if (pFrom == NULL) pFrom=pUnknown;
	while (*pFrom && cntOut) {
		*(pOutPos++) = *(pFrom++);
		cntOut--;
	}
	if (cntOut) { *(pOutPos++) = '\t'; cntOut--; }

    pFrom = data->NAS_id->NAC_address;
	if (pFrom == NULL) pFrom=pUnknown;
	while (*pFrom && cntOut) {
		*(pOutPos++) = *(pFrom++);
		cntOut--;
	}
	if (cntOut) { *(pOutPos++) = '\t'; cntOut--; }

	pFrom = data->NAS_id->username;
	if (pFrom == NULL) pFrom=pUnknown;
	while (*pFrom && cntOut) {
		*(pOutPos++) = *(pFrom++);
		cntOut--;
	}
	if (cntOut) { *(pOutPos++) = '\t'; cntOut--; }

	pFrom = (data->status == TAC_PLUS_AUTHEN_STATUS_PASS) ? "accepted" : "rejected";
	if (pFrom) {
		while (*pFrom && cntOut) {
			*(pOutPos++) = *(pFrom++);
			cntOut--;
		}
	}
	if (cntOut) { *(pOutPos++) = '\t'; cntOut--; }

	pFrom = pUnknown;
	switch (data->service) {
	case TAC_PLUS_AUTHEN_SVC_LOGIN:
		pFrom = "login";
		break;
	case TAC_PLUS_AUTHEN_SVC_ENABLE:
		pFrom = "enable";
		break;
	case TAC_PLUS_AUTHEN_SVC_PPP:
		pFrom = "ppp";
		break;
	case TAC_PLUS_AUTHEN_SVC_ARAP:
		pFrom = "arap";
		break;
	case TAC_PLUS_AUTHEN_SVC_PT:
		pFrom = "pt";
		break;
	case TAC_PLUS_AUTHEN_SVC_RCMD:
		pFrom = "rcmd";
		break;
	case TAC_PLUS_AUTHEN_SVC_X25:
		pFrom = "x25";
		break;
	case TAC_PLUS_AUTHEN_SVC_NASI:
		pFrom = "nasi";
		break;
	}

	while (*pFrom && cntOut) {
		*(pOutPos++) = *(pFrom++);
		cntOut--;
	}
	if (cntOut) { *(pOutPos++) = '\t'; cntOut--; }
	if (cntOut) { *(pOutPos++) = '\n'; cntOut--; }

	logfd = open(filename, O_CREAT | O_WRONLY | O_APPEND,0640);
	if (logfd < 0) {
		report(LOG_ERR, "Can't open acct file %s -- %s", filename, strerror(errno));
		return(0);
	}
	len = strlen(szOut);
    if (write(logfd, szOut, len) != len) {
		report(LOG_ERR, "accesslog: couldn't write file %s %s",
	       filename, strerror(errno));
		return(0);
    }
	close(logfd);

	return(1);
}
