/*
     Verify that this user/password is valid per a database.
     Return 1 if verified, 0 otherwise.
     
     Format of connection string (look like internet URL):

       db://user:password@hostname/database/table?name&passwd

     Example connect to Oracle RDBMS at user 'roger' with password
     'tiger', to 'host.domain.com' database server, fields name in
     table is 'name' and 'passwd' in 'oshadow' table:

       oracle://roger:tiger@host.domain.com/tacacs/oshadow?name&passwd

     DONE:
     12-nov-1998 Created
                 Add DB support to 'login = db <string>'
     14-nov-1998 Change Tacacs+ version from 0.95 to 3.0.9
     18-nov-1998 Added code for Oracle [version 8.0.5]
     27-nov-1998 Tested with 30'000 usernames Oracle database
                 Added DB support to global configuration
                 'default authentication = db <string>'
     28-nov-1998 Added code for NULL database %)
     
     FUTURE:
     Make *_db_verify() the functions is reenterable
     More security for connection to database
     GDBM support
     Separate debug logging
     Perfomance testing on 10000 records in Oracle database
     (in guide sayd about 3 auth/sec on Ultra 2 - hmm)
     
     -------------------------------------------------------
     fil@artelecom.ru                   http://twister.pp.ru
 
 ****************************************************************************
				    PART II

   I am added some extra extension. Like MySQL and PostgreSQL database support
   And change most of lines for use dynamic memory allocation. db_accounting 
   added by me.
  
   devrim(devrim@gazi.edu.tr)

 ****************************************************************************
				    PART III

   I have change the way how the DB flows to address features such as recursive
   lookups for users and groups.  Plus, I added and changed a few things.

   August 20, 2002 - Added function (parse_db) and structure (dbinfo) to
   		     minimize the amount of parameters that are sent to the
		     other functions.

   August 22, 2002 - Changed db_acct to allow for more information to be
                     inserted into the database

   August 22, 2002 - Changed db_acct to create the SQL statement then send
   		     it the supporting functions.  All databases use in this
		     program support SQL except for db_null.

   baram01@hotmail.com
*/

#include "tac_plus.h"

#if defined(USE_DB)
#include <stdio.h>

#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#include "db.h"

struct author_user {
	int svc_dflt;
	char username[30];
	char member[30];
	char pre_author[30];
	char post_author[30];
};

/* For checking DB type */
int 
check_db_type( char *db_type )
{
char *dbp[]=DEFINED_DB;
int ret=1,i;

for (i=0; dbp[i] ; i++ ) {
	if(!strcmp(db_type,dbp[i])) {
		ret=0;
		break;
	}
}
return ret;
}

TAC_DBINFO *
db_parse_url(char *url)
{
	char *pUrl;
	void *pAlloc;
	TAC_DBINFO *db_info;
	int	len;

	if (!url) {
		report(LOG_ERR,"db_parse_url: called with NULL url");
		return(NULL);
	}

	if (debug & DEBUG_PARSE_FLAG) {
		report(LOG_DEBUG, "db_parse_url: %s", url);
	}

	len = strlen(url) + sizeof(TAC_DBINFO) + 1;

	len *= sizeof(char);	/* to make sure there is enough room */

	pAlloc = tac_malloc(len);  /* don't need to check return because tac_malloc does */
	bzero(pAlloc, len);

	db_info = (TAC_DBINFO *)pAlloc;
	pUrl = pAlloc;
	pUrl += sizeof(TAC_DBINFO);	  /* put the copy of the string after the structure */

	strcpy(pUrl, url);
	
	db_info->type = pUrl;
	
    if( NULL == (pUrl = (char *)strstr( pUrl, "://" )) ) {
		report(LOG_ERR,"db_parse_url: parsing user");
		free(pAlloc);
		return(NULL);
	}
	*pUrl = '\0'; 
	pUrl += 3;

	/* For recognize db authentication database */
    
    if (check_db_type(db_info->type)) {
		report(LOG_ERR,"db_parse_url: %s DB authentication scheme " \
			   "not recognized by tac_plus",db_info->type);
		free(pAlloc);
		return(NULL);
	}

	db_info->user = pUrl;

	if( NULL == (pUrl = (char *)strchr( pUrl, ':' )) ) {
		report(LOG_ERR,"db_parse_url: parsing password");
		free(pAlloc);
		return(NULL);
	}
	*(pUrl++) = '\0';
	db_info->password = pUrl;
    
    if( NULL == (pUrl = (char *)strchr( pUrl, '@' )) ) {
		report(LOG_ERR,"db_parse_url: parsing hostname");
		free(pAlloc);
		return(NULL);
    }
	*(pUrl++) = '\0';
	db_info->host = pUrl;
    
	if( NULL == (pUrl = (char *)strchr( pUrl, '/' )) ) {
		report(LOG_ERR,"db_parse_url: parsing database name");
		free(pAlloc);
		return(NULL);
	} 
	*(pUrl++) = '\0';
	db_info->database = pUrl;
    
	if( NULL == (pUrl = (char *)strchr( pUrl, '/' )) ) {
		db_info->table = DEFAULT_DB_TABLE_NAME;
		db_info->field_name = DEFAULT_DB_FIELD_NAME_USER;
		db_info->field_passwd = DEFAULT_DB_FIELD_NAME_PASSWD;
		if (debug & DEBUG_PARSE_FLAG)
			report(LOG_INFO,"db_parse_url: Using default table");
	} else { 
		*(pUrl++) = '\0';
		db_info->table = pUrl;

		if( NULL == (pUrl = (char *)strchr( pUrl, '?' )) ) {
			db_info->field_name = DEFAULT_DB_FIELD_NAME_USER;
			db_info->field_passwd = DEFAULT_DB_FIELD_NAME_PASSWD;
			if (debug & DEBUG_PARSE_FLAG)
				report(LOG_DEBUG, "Using default field names: " \
					   "userid: %s, password: %s",
					   db_info->field_name, db_info->field_passwd);
		} else {
			*(pUrl++) = '\0';
			db_info->field_name = pUrl;

			if( NULL == (pUrl = (char *)strchr( pUrl, '&' )) ) {
				db_info->field_passwd = DEFAULT_DB_FIELD_NAME_PASSWD;
				if (debug & DEBUG_PARSE_FLAG)
					report(LOG_DEBUG, "Using default password field name %s",
						   db_info->field_passwd );
			} else {
				*(pUrl++) = '\0';
				db_info->field_passwd = pUrl;
			}
		}
	}

    /* Parse database connection string */
	if (debug & DEBUG_PARSE_FLAG)
		report(LOG_DEBUG, "db_parse_url: db_type=%s, db_user=%s, " \
			   "db_password=%s db_hostname=%s, db_name=%s, " \
			   "db_table=%s, dbfield_name=%s, dbfield_passwd=%s", 
			   db_info->type, db_info->user, db_info->password, 
			   db_info->host, db_info->database, db_info->table, 
			   db_info->field_name, db_info->field_passwd);

	return (db_info);
}

int db_connect(DBH *dbh, TAC_DBINFO *dbinfo)
{
#if defined(DB_MYSQL)
	if (!mysql_init(dbh)) {
		if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_ACCT_FLAG|DEBUG_PASSWD_FLAG))
			report(LOG_DEBUG, "db_connect: Cannot initialize");
		return (0);
	}
	if (!mysql_real_connect(dbh, dbinfo->host, dbinfo->user, dbinfo->password, dbinfo->database, 0, NULL, 0))
	{
		if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG|DEBUG_ACCT_FLAG))
			report(LOG_DEBUG, "db_connect: Cannot connect to %s database %s. Error: %s", dbinfo->host, dbinfo->database, mysql_error(dbh));

		/** Need add for redundancy here **/
		return (0);
	}
#endif

#if defined(DB_PGSQL)
	dbh->conn = PQsetdbLogin(dbinfo->host, NULL, NULL, NULL, dbinfo->database, dbinfo->user, dbinfo->password);
	if (PQstatus(dbh->conn) == CONNECTION_BAD) {
		if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG|DEBUG_ACCT_FLAG))
			report(LOG_DEBUG, "db_connect: Cannot connect to %s database %s. Error: %s", dbinfo->host, dbinfo->database, PQerrorMessage(dbh));
		return (0);
	}
#endif

	return (1);
}

void db_close(DBH *dbh)
{
#if defined(DB_MYSQL)
	mysql_close(dbh);
#endif

#if defined(DB_PGSQL)
	PQfinish(dbh->conn);
#endif
}

DB_RES *db_store_result(DBH *dbh)
{
#if defined(DB_MYSQL)
	return mysql_store_result(dbh);
#endif

#if defined(DB_PGSQL)
	return dbh->res;
#endif
}

int db_num_rows(DB_RES *result)
{
#if defined(DB_MYSQL)
	return mysql_num_rows(result);
#endif

#if defined(DB_PGSQL)
	return PQntuples(result);
#endif
}

DB_ROW db_fetch_row(DB_RES *result)
{
#if defined(DB_MYSQL)
	return mysql_fetch_row(result);
#endif
}

void db_free_result(DB_RES *result)
{
#if defined(DB_MYSQL)
	mysql_free_result(result);
#endif

#if defined(DB_PGSQL)
	PQclear(result);
#endif
}

int db_affected_rows(DBH *dbh)
{
#if defined(DB_MYSQL)
	return mysql_affected_rows(dbh);
#endif
}

#ifdef __STDC__
int db_query(DBH *dbh, char *querystmt,...)
#else
int db_query(dbh, querystmt, va_alist)
DBH *dbh;
char *querystmt;
va_dcl
#endif
{
	char *p, *sval, *sqlcmd, *bufp;
	char digits[16];
	int m, len = 0, ret = 1;
	va_list ap;

	bufp = sqlcmd = (char*)malloc(SQLCMDL);

#ifdef __STDC__
	va_start(ap, querystmt);
#else
	va_start(ap);
#endif

	for (p = querystmt; *p; p++) {
		if (*p != '%') {
			*bufp++ = *p;
			len++;
			continue;
		}
		switch (*++p) {
		case 's':
			sval = va_arg(ap, char *);
			m = strlen(sval);
			break;

		case 'u':
			sprintf(digits, "%u", va_arg(ap, uint));
			m = strlen(digits);
			sval = digits;
			break;

		case 'x':
			sprintf(digits, "%x", va_arg(ap, uint));
			m = strlen(digits);
			sval = digits;
			break;

		case 'd':
			sprintf(digits, "%d", va_arg(ap, int));
			m = strlen(digits);
			sval = digits;
			break;
		}

		memcpy(bufp, sval, m);
		bufp += m;
		len += m;
		continue;
	}

	sqlcmd[len] = '\0';
	va_end(ap);
	
#if defined(DB_MYSQL)
	if (mysql_real_query(dbh, sqlcmd, len)) {
		if (debug)
			report(LOG_DEBUG, "db_query: Cannot query using MySQL. Error - %s", mysql_error(dbh));
		ret = 0;
	}
#endif

#if defined(DB_PGSQL)
	dbh->res = PQexec(dbh->conn, sqlcmd);
	if (!dbh->res || PQresultStatus(dbh->res) != PGRES_COMMAND_OK )
	{
		if (debug)
			report(LOG_DEBUG, "db_query: Cannot query using PGSQL.  Error - %s", PQerrorMessage(dbh->conn));
		ret = 0;
	}
#endif

	free(sqlcmd);
	return(ret);
}

int db_get_host(char *field, tac_uint32 nas, char *data, TAC_DBINFO *dbinfo)
{
	int ret = 1;
	DBH dbh;
	DB_RES *res;
	DB_ROW row;
	
	if (!field || !data || !dbinfo) {  /* no sense in continuing if we have NULL input */
		return(2);
	}

	if (debug)
		report(LOG_DEBUG, "db_get_host: getting %s from nas(%s)", field, 
			   cfg_addr_as_string(nas));

	if (db_connect(&dbh, dbinfo)) {
		if (db_query(&dbh, HOSTAUTHSQL, field, nas)) {
			res = db_store_result(&dbh);
			if (db_num_rows(res) >= 1) {
				while (ret && (row = db_fetch_row(res))) {
					if (!tac_empty_string(row[0])) {
						strcpy(data, row[0]);
						if (debug)
							report(LOG_DEBUG, "db_get_host: got (%s)", data);
						ret = 0;
					}
				}
			}
			db_free_result(res);
		}
		db_close(&dbh);
	}
/*	
	if (dbinfo->type) {
#if defined(DB_MYSQL)
		if (!strcmp(dbinfo->type, "mysql"))
			ret = mysql_db_get_host( field, nas, data, dbinfo );
#endif
	}
*/
	return ret;
}

int
db_verify(char *user, tac_uint32 host, char *clientip, char *users_passwd, TAC_DBINFO *dbinfo, int recurse)
{
   int cont = 1, logintype = 0, acl;
   char member[30], password[40];
   DBH dbh;
   DB_RES *res;
   DB_ROW row;

   if (!dbinfo) {
	   report(LOG_ERR, "db_verify: called with NULL db_info");
	   return(0);
   }

   /* Check for empty userid or password */
   if (user && users_passwd) {
	   if (!*user || !*users_passwd) {
		   if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
			  report(LOG_DEBUG, "db_verify: called with empty userid or password");
		   return(0);
	   }
   } else {
	   if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
		  report(LOG_DEBUG, "db_verify: called with missing userid or password");
	   return(0);
   }

   if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
	   report(LOG_DEBUG, "db_verify: verify user %s to %s database", 
			  user, dbinfo->type);

   /* Check for empty passwords */
   if (dbinfo->user && dbinfo->password ) {
	   if (!*(dbinfo->user) || !*(dbinfo->password)) {
		   if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
			   report(LOG_ERR, "db_verify: Empty database userid or password");
		   return (0);
	   }
   } else {
	   if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
		   report(LOG_DEBUG, "db_verify: Missing database userid or password");
	   return (0);
   }

   if (!db_connect(&dbh, dbinfo)) {
	   if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
		   report(LOG_DEBUG, "db_verify:  Cannot connected to the databases");
	   return (0);
   }

   /** find a better way to detect type of login **/
   if (strcmp(dbinfo->field_passwd, "enable")==0)
	   logintype=1;

   if (host) {
	   if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
		   report(LOG_DEBUG, "db_verify:  verifying host(%u)", host);
	   /** This function will not be supported in the future **/
	   if (logintype && db_query(&dbh, HOSTAUTHSQL, dbinfo->field_passwd, host)) {
		   res = db_store_result(&dbh);
		   if (db_num_rows(res)) {
			   row = db_fetch_row(res);
			   strcpy(password, row[0]);
			   db_free_result(res);
			   if (!tac_empty_string(password))
				   goto verify;
		   }
		   db_free_result(res);
	   }
   }

   if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
	   report(LOG_DEBUG,"db_verify: verify user(%s)", user);

   if (!db_query(&dbh, AUTHSQL, dbinfo->field_passwd, dbinfo->table, dbinfo->field_name, user)) {
	   if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
		   report(LOG_DEBUG, "db_verify: database problems");
	   db_close(&dbh);
	   return(0);
   }

   res = db_store_result(&dbh);
   if (db_num_rows(res) == 0) {
	   if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
		   report(LOG_DEBUG, "db_verify: user(%s) not found", user);
	   db_close(&dbh);
	   return(0);
   }
   row = db_fetch_row(res);
   strcpy(password, row[0]);
   strcpy(member, row[1]);
   db_free_result(res);

   /** Verify if user is allowed to access the NAS **/
   if (host) {
   	if (db_query(&dbh, HOSTACLSQL,logintype?"enableacl":"loginacl", host)) {
		res = db_store_result(&dbh);
		if (db_num_rows(res) > 0) {
			row = db_fetch_row(res);
			acl = atoi(row[0]);

			/** checking to see if not implicit allow **/
			if (acl) {
				db_free_result(res);
				if (db_query(&dbh, NASACLSQL, acl, user, member)) {
					res = db_store_result(&dbh);
					if (db_num_rows(res) == 0) {
						if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
							report(LOG_DEBUG, "db_verify: user(%s) is not allowed to access nas(%u)", user, host);
						db_free_result(res);
						db_close(&dbh);
						return (0);
					}

					db_free_result(res);
					if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
						report(LOG_DEBUG, "db_verify: user(%s) is allowed to access nas(%u)", user, host);
				}
			} else {
				if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
					report(LOG_DEBUG, "db_verify: user(%s) is implicitly allowed to nas(%u)", user, host);
			}
		} else {
			if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
				report(LOG_DEBUG, "db_verify: user(%s) do not have permission to access this nas(%u)", user, host);
			db_free_result(res);
			db_close(&dbh);
			return (0);
		}
	} else {
	   if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
		   report(LOG_DEBUG, "db_verify: problems with HOSTACL query");
	}
   }

   /** Verify if user is allowed to access the NAS from current IP **/
   if (!tac_empty_string(clientip)) {
	   if (!db_query(&dbh, USERACCESSSQL, user, member, clientip)) {
		   if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
			   report(LOG_DEBUG, "db_verify: database problems.");
		   db_close(&dbh);
		   return(0);
	   }
	   res = db_store_result(&dbh);
	   if (db_num_rows(res)) {
		   while (cont && (row = db_fetch_row(res))) {
			   if (strcmp(row[0],"58")==0) {
				   if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
					   report(LOG_DEBUG,"db_verify: User is not authorized to access nas(%u) from this address(%s)", host, clientip);
				   db_free_result(res);
				   db_close(&dbh);
				   return(0);
			   } else if (strcmp(row[0],"57")==0)
				   cont=0;
		   }
		   if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
			   report(LOG_DEBUG, "db_verify: User is authorized to access nas(%u) from this address(%s)", host, clientip);
	   }
   } else {
	   if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
		   report(LOG_DEBUG, "db_verify: User is login in through console");
   }

   /** Continue with regular processing **/
   if (tac_empty_string(password)) {
	   if (!recurse) {
		   if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
			   report(LOG_DEBUG, "db_verify: recurse was not requested.");
		   db_close(&dbh);
		   return(0);
	   }

	   if (tac_empty_string(member)) {
		   if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
			   report(LOG_DEBUG, "db_verify: User(%s) is not a member of any group.", user);
		   db_close(&dbh);
		   return(0);
	   }

	   if (!db_query(&dbh, AUTHSQL, dbinfo->field_passwd, dbinfo->table, dbinfo->field_name, member)) {
		   db_close(&dbh);
		   return(0);
	   }

	   res = db_store_result(&dbh);
	   if (db_num_rows(res) == 0) {
		   if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
			   report(LOG_DEBUG, "db_verify: member(%s) not found", member);
		   db_close(&dbh);
		   return(0);
	   }
	   row = db_fetch_row(res);
	   strcpy(password, row[0]);
	   db_free_result(res);

	   if (tac_empty_string(password)) {
		   db_close(&dbh);
	   }
   }

verify:
   db_close(&dbh);

   /** Verify password **/
   if (!des_verify(users_passwd, password))
	   return (0);

    return (1); 
}

int
db_access( AUTHEN_DATA *data )
{
	int ret = 0;
	char *uid, *nas, *client_ip, *terminal, *status, *service;

	DBH dbh;
	DB_RES *res;
	DB_ROW row;

	if (!data) return ret;

	/** Point default db to the correct table for logging **/
	if ((uid=data->NAS_id->username) == NULL) {
		uid = "unknown";
	}

	if ((nas=data->NAS_id->NAS_name) == NULL) {
		nas = "unknown";
	}

	if ((client_ip=data->NAS_id->NAC_address) == NULL) {
		client_ip = "unknown";
	}

	if ((terminal=data->NAS_id->NAS_port) == NULL) {
		terminal = "unknown";
	}

	status = (data->status == TAC_PLUS_AUTHEN_STATUS_PASS) ? "accepted" : "rejected";

	switch (data->service) {
		case TAC_PLUS_AUTHEN_SVC_LOGIN:
			service = "login";
			break;
		case TAC_PLUS_AUTHEN_SVC_ENABLE:
			service = "enable";
			break;
		case TAC_PLUS_AUTHEN_SVC_PPP:
			service = "ppp";
			break;
		case TAC_PLUS_AUTHEN_SVC_ARAP:
			service = "arap";
			break;
		case TAC_PLUS_AUTHEN_SVC_PT:
			service = "pt";
			break;
		case TAC_PLUS_AUTHEN_SVC_RCMD:
			service = "rcmd";
			break;
		case TAC_PLUS_AUTHEN_SVC_X25:
			service = "x25";
			break;
		case TAC_PLUS_AUTHEN_SVC_NASI:
			service = "nasi";
			break;
		default:
			service = "unknown";
	}

	if (db_connect(&dbh, session.db_default)) {
		if (db_query(&dbh, ACCESSSQL, nas, terminal, uid, client_ip, service, status)) {
			if (db_affected_rows(&dbh)>0) {
				ret = 1;

				if (debug & DEBUG_ACCT_FLAG)
					report(LOG_DEBUG, "db_access: inserting record is successfull");
			}
		}
	}

	db_close(&dbh);
	return (ret);
}

int
db_acct( struct acct_rec *rec, TAC_DBINFO *dbinfo )
{
    int ret = 0;
    char *uid, *nas, *client_ip, *terminal, *type, *service, *priv_lvl, *cmd,
	 *elapsed_time, *bytes_in, *bytes_out;
    char *pUnknown = "unknown";

	DBH dbh;
	DB_RES *res;
	DB_ROW row;

	if (!dbinfo) {
		report(LOG_ERR, "db_acct: called with NULL db_info");
		return(0);
	}

	if (debug & DEBUG_ACCT_FLAG) 
		report(LOG_DEBUG,"db_acct: log accounting record to database");

    /* Check for empty passwords */
	if (dbinfo->user && dbinfo->password ) {
		if (!*(dbinfo->user) || !*(dbinfo->password)) {
			if (debug & DEBUG_ACCT_FLAG)
				report(LOG_ERR, "db_acct: Empty database userid or password");
			return (0);
		}
    } else {
		if (debug & DEBUG_ACCT_FLAG)
			report(LOG_DEBUG, "db_acct: Missing database userid or password");
		return (0);
	}

/* Find some attributes  for accounting */
    if ((uid=rec->identity->username) == NULL) {
        if (debug & DEBUG_ACCT_FLAG) 
			report(LOG_DEBUG,"db_acct: Can't find username!");
		uid = pUnknown;
	}
	if ((nas=rec->identity->NAS_name)==NULL) {
		if (debug & DEBUG_ACCT_FLAG) 
			report(LOG_DEBUG,"db_acct: Can't find NAS name!");
		nas = pUnknown;
	}
    if ((client_ip=rec->identity->NAC_address)==NULL) {
        	if (debug & DEBUG_ACCT_FLAG) 
			report(LOG_DEBUG,"db_acct: Can't find client address!");
		client_ip = pUnknown;
	}
    if ((terminal=rec->identity->NAS_port)==NULL) {
        	if (debug & DEBUG_ACCT_FLAG) 
			report(LOG_DEBUG,"db_acct: Can't find terminal!");
		terminal = pUnknown;
	}
    switch(rec->acct_type) {
    case ACCT_TYPE_UPDATE:
        type="update";
        break;
    case ACCT_TYPE_START:
        type="start";
        break;
    case ACCT_TYPE_STOP:
        type="stop";
        break;
    default:
        type=pUnknown;
        break;
    }
    if ((service=find_attr_value("service", rec->args, rec->num_args))==NULL) {
		service="NULL";
	}
    if ((priv_lvl=find_attr_value("priv_lvl", rec->args, rec->num_args))==NULL) {
		priv_lvl="NULL";
	}
    if ((elapsed_time=find_attr_value("elapsed_time", rec->args, rec->num_args))==NULL) {
		elapsed_time="NULL";
	}
    if ((bytes_in=find_attr_value("bytes_in", rec->args, rec->num_args))==NULL)
    	bytes_in="NULL";
    if ((bytes_out=find_attr_value("bytes_out",rec->args, rec->num_args))==NULL)
    	bytes_out="NULL";
    if ((cmd=find_attr_value("cmd", rec->args, rec->num_args))==NULL)
	cmd="";

	if (db_connect(&dbh, dbinfo)) {
		if (db_query(&dbh, ACCTSQL, dbinfo->table, nas, uid, terminal, client_ip, type, service, priv_lvl, elapsed_time, bytes_in, bytes_out, cmd)) {
			if (db_affected_rows(&dbh)>0) {
				ret = 1;

				if (debug & DEBUG_ACCT_FLAG)
					report(LOG_DEBUG, "db_acct: inserting record is successfull");
			}
		}
	}

	/** Need to consolidate this information in the main table --ARY **/
    	if ((!strcasecmp(service,"shell"))&&(!session.db_default))
		ret = mysql_db_shell_acct(dbinfo, nas, client_ip, uid, rec);

    db_close(&dbh);
    return (ret);

}

int db_author( struct author_data *data )
{
	int ret = 1, svc;
	DBH dbh;
	DB_RES *res;
	DB_ROW row;
	char *cmd, *protocol, *svcname;
	struct author_user user;

	data->status = AUTHOR_STATUS_FAIL;
	data->output_args = NULL;
	data->num_out_args = 0;

	if (debug & DEBUG_AUTHOR_FLAG)
		report(LOG_DEBUG, "db_author: authorizing user(%s)", data->id->username);

	strcpy(user.username, data->id->username);
	protocol = NULL;

	if (db_connect(&dbh, session.db_default)) {
		if (db_query(&dbh, "SELECT b_author, a_author, svc_dflt, gid FROM %s where uid = '%s'", session.db_default->table, user.username)) {
			res = db_store_result(&dbh);
			if (db_num_rows(res) > 0) {
				row = db_fetch_row(res);
				strcpy(user.member, row[3]?row[3]:"");
				strcpy(user.pre_author, row[0]?row[0]:"");
				strcpy(user.post_author, row[1]?row[1]:"");
				user.svc_dflt = atoi(row[2]?row[2]:"-1");
				db_free_result(res);

				if ( pre_authorization(user.username, user.pre_author[0]?user.pre_author:NULL, data)) {
					db_close(&dbh);
					return (0);
				}

				if (!(svc=get_nas_svc(data, &cmd, &protocol, &svcname))) {
					data->status = AUTHOR_STATUS_ERROR;
					data->admin_msg = tac_strdup("No identifiable service/protocol in authorization request");
					if (debug & DEBUG_AUTHOR_FLAG)
						report(LOG_DEBUG, "db_author: %s for %s", data->admin_msg, user.username);
					db_close(&dbh);
					return (0);
				}

				switch (svc) {
					default:
						report(LOG_ERR, "db_author: bad service type (%d) from %s", svc, session.peer);
						break;

					case N_svc_cmd:
						ret = db_authorize_cmd(&dbh, &user, cmd, data);
						break;

					case N_svc_exec:
						if (db_authorize_exec(&dbh, &user, data))
								return (0);
						
					case N_svc_arap:
					case N_svc_ppp:
					case N_svc_slip:
						ret = db_authorize_svc(&dbh, &user, svc, protocol, NULL, data);
						break;

					case N_svc:
						ret = db_authorize_svc(&dbh, &user, svc, protocol, svcname, data);
						break;
				}

				post_authorization(user.username, user.post_author[0]?user.post_author:NULL, data);
			}
			db_free_result(res);
		}
		db_close(&dbh);
	}
	return (ret);
}

static char ** get_svc_attrs(DB_RES *res, int *denyp)
{
	DB_ROW row;
	char **args;
	int i = db_num_rows(res);

	*denyp = 1;
	args = (char**) tac_malloc(sizeof(char *) * (i+1));
	i = 0;
	while (row = db_fetch_row(res)) {
		char *arg = tac_strdup(row[1]);
		char *p = index(arg, '=');

		if (p && atoi(row[0]) == N_optarg)
			*p = '*';
		args[i++] = arg;
	}
	args[i] = NULL;
	return (args);
}

int get_svc_node(DBH *dbh, char *uid, int svc, char *value, char *value1)
{
	switch (svc) {
		case N_svc:
			if (!db_query(dbh, "SELECT type, value FROM node WHERE uid = '%s' AND service = %d AND value1 ='%s'", uid, svc, value1))
				return (1);
			break;

		case N_svc_ppp:
			if (!db_query(dbh, "SELECT type, value FROM node WHERE uid = '%s' AND service = %d AND value1 = '%s'", uid, svc, value1) )
				return (1);
			break;
			
	        case N_svc_cmd:
		       if (!db_query(dbh, "SELECT uid, type, value1 FROM node WHERE uid = '%s' and service = %d and value = '%s' and '%s' REGEXP value1 ORDER BY seq" , uid, svc, value, value1))
		       return (1);
		       break;
		default:
		       if (!db_query(dbh, "SELECT type, value FROM node WHERE uid = '%s' AND service = %d", uid, svc) )
		                return (1);
	}

	return (0);
}

int db_authorize_svc(dbh, user, svc, protocol, svcname, data)
DBH *dbh;
struct author_user *user;
int svc;
char *protocol, *svcname;
struct author_data *data;
{
	DB_RES *res;
	char **cfg_args;
	int deny_by_default;

	if (debug & DEBUG_AUTHOR_FLAG) {
		report(LOG_DEBUG, "db_authorize_svc: service authorization for %s", user->username);
	}

	if (get_svc_node(dbh, user->username, svc, "", protocol)) {
		data->status = AUTHOR_STATUS_ERROR;
		data->num_out_args = 0;
		return (1);
	}

	res = db_store_result(dbh);
	if (db_num_rows(res) <= 0) {
		db_free_result(res);

		if (!user->member) goto NEXT;
		if (debug & DEBUG_AUTHOR_FLAG)
			report(LOG_DEBUG, "db_authorise_svc: service authorization for %s", user->member);

		if (get_svc_node(dbh, user->member, svc, "", svcname)) {
			data->status = AUTHOR_STATUS_ERROR;
			data->num_out_args = 0;
			return (1);
		}

		res = db_store_result(dbh);
		if (db_num_rows(res) >= 1)
			goto NEXT2;
		db_free_result(res);
NEXT:
		if (user->svc_dflt ==  1) {
			if (debug & DEBUG_AUTHOR_FLAG)
				report(LOG_DEBUG,"db_authorize_svc: svc=%s protocol=%s svcname=%s not found, permitted by default",
					cfg_nodestring(svc), protocol ? protocol : "",
					svcname ? svcname : ""); 

			data->status = AUTHOR_STATUS_PASS_ADD;
			data->num_out_args = 0;
			data->output_args = NULL;
			return (0);
		}

		if (debug & DEBUG_AUTHOR_FLAG)
				report(LOG_DEBUG,"db_authorize_svc: svc=%s protocol=%s not found, denied by default",
					cfg_nodestring(svc), protocol ? protocol : ""); 

		data->status = AUTHOR_STATUS_FAIL;
		data->output_args = NULL;
		data->num_out_args = 0;
		return (0);
	}
NEXT2:
	cfg_args = get_svc_attrs( res, &deny_by_default);
	return (author_svc(cfg_args, deny_by_default, data));
}

int db_authorize_cmd(dbh, user, cmd, data)
DBH *dbh;
struct author_user *user;
char *cmd;
struct author_data *data;
{
	DB_RES *res;
	DB_ROW row;
	char *args;

	data->num_out_args = 0;

	args = assemble_args(data);
	data->num_out_args = 0;
	if (debug & DEBUG_AUTHOR_FLAG)
		report(LOG_DEBUG, "db_authorize_cmd: command (%s %s) authorization for %s", cmd?cmd:"", args?args:"", user->username);


	if (!cmd) {
		data->status = AUTHOR_STATUS_ERROR;
		data->admin_msg = tac_strdup("No command found");
		report(LOG_ERR, "db_authorize_cmd: %s - %s %s", session.peer, cmd, data->admin_msg);
		if (args) free(args);
		return (0);
	}

	if (get_svc_node(dbh, user->username, N_svc_cmd, cmd, args?args:"")) {
		data->status = AUTHOR_STATUS_ERROR;
		if (args) free(args);
		return (1);
	}
	res = db_store_result(dbh);
	if (db_num_rows(res) <= 0) {
		db_free_result(res);

		if (!user->member) goto next1;
		if (debug & DEBUG_AUTHOR_FLAG)
			report(LOG_DEBUG, "db_authorize_cmd: command (%s %s) authorization for %s", cmd?cmd:"", args?args:"", user->member);
		if (get_svc_node(dbh, user->member, N_svc_cmd, cmd, args?args:"")) {
			data->status = AUTHOR_STATUS_ERROR;
			if (args) free(args);
			return (1);
		}
		res = db_store_result(dbh);
		if (db_num_rows(res) >= 1)
			goto next2;

next1:
		if (args) free(args);

                if (user->svc_dflt ==  1) {
                        if (debug & DEBUG_AUTHOR_FLAG)
                                report(LOG_DEBUG,"db_authorize_cmd: command(%s) not found, permitted by default", cmd);
                        data->status = AUTHOR_STATUS_PASS_ADD;
                        return (0);
                }

                if (debug & DEBUG_AUTHOR_FLAG)
			report(LOG_DEBUG,"db_authorize_cmd: command(%s) not found, denied by default", cmd);
                data->status = AUTHOR_STATUS_FAIL;
                return (0);
	}

	/** command exists.  check if command is DENIED or PERMITTED **/
next2:
	row = db_fetch_row(res);
	switch (atoi(row[1])) {
	case N_permit:
		if (debug & DEBUG_AUTHOR_FLAG) 
			report(LOG_DEBUG, "%s %s permitted", cmd, args);
		data->status = AUTHOR_STATUS_PASS_ADD;
		break;
	case N_deny:
		if (debug & DEBUG_AUTHOR_FLAG)
			report(LOG_DEBUG, "%s %s denied", cmd, args);
		data->status = AUTHOR_STATUS_FAIL;
		break;
	default:
		data->status = AUTHOR_STATUS_ERROR;
		data->admin_msg = tac_strdup("Server error illegal configuration");
		report(LOG_ERR, "%s: %s %s %s", session.peer, cmd, args,
			data->admin_msg);
		break;
	}

	if (args) free(args);
	db_free_result(res);
	return (0);
}

int
db_authorize_exec(dbh, user, data)
DBH *dbh;
struct author_user *user;
struct author_data *data;
{
	DB_RES *res;
	int num_rows;

	if (debug & DEBUG_AUTHOR_FLAG)
		report(LOG_DEBUG, "db_authorize_exec: exec authorization for %s", data->id->username);

	if (user->svc_dflt == 1) {
		if (debug & DEBUG_AUTHOR_FLAG)
			report(LOG_DEBUG, "db_author: exec permitted by default");
		data->status = AUTHOR_STATUS_PASS_ADD;
		data->output_args = NULL;
		data->num_out_args = 0;
		return(1);
	}

	if (user->member) {
		if (debug & DEBUG_AUTHOR_FLAG)
			report(LOG_DEBUG, "db_authorize_exec: %s is part of a group. Authorizing both user and group (%s) for exec.", data->id->username, user->member);
		if (!db_query(dbh, "SELECT service FROM node WHERE uid IN ('%s','%s') and service = %d" , user->username, user->member, N_svc_exec)) {
			data->status = AUTHOR_STATUS_ERROR;
			data->output_args = NULL;
			data->num_out_args = 0;
			return (1);
		}
	} else {
		if (debug & DEBUG_AUTHOR_FLAG)
			report(LOG_DEBUG, "db_authorize_exec: %s is not part of a group. Authorizing user for exec.", data->id->username);

		if (!db_query(dbh, "SELECT service FROM node WHERE uid = '%s' and service = %d" , user->username, N_svc_exec)) {
			data->status = AUTHOR_STATUS_ERROR;
			data->output_args = NULL;
			data->num_out_args = 0;
			return (1);
		}
	}
	res = db_store_result(dbh);
	num_rows = db_num_rows(res);
	if (num_rows > 0) {
		if (debug & DEBUG_AUTHOR_FLAG)
			report(LOG_DEBUG, "db_authorize_exec: exec is explicitly permitted");
		db_free_result(res);
		return (0);
	}
	db_free_result(res);

	if  (user->member) {
		if (debug & DEBUG_AUTHOR_FLAG)
			report(LOG_DEBUG, "db_authorize_exec: %s is part of a group. Authorizing user and group for commands.", data->id->username);
		if (!db_query(dbh, "SELECT service FROM node WHERE uid IN ('%s', '%s') and service = %d" , user->username, user->member, N_svc_cmd)) {
			data->status = AUTHOR_STATUS_ERROR;
			data->output_args = NULL;
			data->num_out_args = 0;
			return (1);
		}
	} else {
		if (debug & DEBUG_AUTHOR_FLAG)
			report(LOG_DEBUG, "db_authorize_exec: %s is not part of a group. Authorizing user for commands.", data->id->username);
		if (!db_query(dbh, "SELECT service FROM node WHERE uid = '%s' and service = %d" , user->username, N_svc_cmd)) {
			data->status = AUTHOR_STATUS_ERROR;
			data->output_args = NULL;
			data->num_out_args = 0;
			return (1);
		}
	}
	res = db_store_result(dbh);
	if (db_num_rows(res) > 0) {
		if (debug & DEBUG_AUTHOR_FLAG)
			report(LOG_DEBUG, "db_authorize_exec: commands configured thus exec permitted");
		data->status = AUTHOR_STATUS_PASS_ADD;
		data->output_args = NULL;
		data->num_out_args = 0;
		return (1);
	}
	db_free_result(res);

	if (debug & DEBUG_AUTHOR_FLAG)
		report(LOG_DEBUG, "db_authorize_exec: exec denied by default");

	data->status = AUTHOR_STATUS_FAIL;
	data->output_args = NULL;
	data->num_out_args = 0;
	return (1);
}

#endif /* DB */

