/*

		Writen by Devrim SERAL(devrim@gazi.edu.tr)
     This program writen for MySQL Authentication and Accounting Propose

License: This code is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later version.

*/

#include "tac_plus.h"

#if defined(DB_MYSQL) && defined(USE_DB)

#include <stdio.h>
#include <mysql.h>
#include "db.h"

#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif


int mysql_db_connect(mysqldb, dbinfo)
MYSQL *mysqldb;
TAC_DBINFO *dbinfo;
{
/* Connect to database server */

    if ( !mysql_init(mysqldb) )
	{
        if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_ACCT_FLAG))
			report(LOG_DEBUG, "mysql_connect: cannot initialize");
		return(0);
	}

    if (!mysql_real_connect(mysqldb,dbinfo->host,dbinfo->user,dbinfo->password,dbinfo->database, 0, NULL, 0))
	{
		if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_ACCT_FLAG))
			report(LOG_DEBUG, "mysql_connect: cannot connect as %s. Error: %s", dbinfo->user, mysql_error(mysqldb));
		return(0);
	}

	return (1);
}

/**
#ifdef __STDC__
int mysql_db_query(MYSQL *mysqldb, char *querystmt,...)
#else
int mysql_db_query(mysqldb, querystmt, va_alist)
MYSQL *mysqldb;
char *querystmt;
va_dcl
#endif
{
	char *p, *sval, *mysqlcmd, *bufp;
	char digits[16];
	int m, len = 0;
	va_list ap;

	bufp = mysqlcmd = (char*)malloc(SQLCMDL);

#ifdef __STDC__
	va_start(ap, querystmt);
#else
	va_start(ap);
#endif

	for (p = querystmt; *p; p++) {
		if (*p != '%') {
			*bufp++ = *p;
			len++;
			continue;
		}
		switch (*++p) {
		case 's':
			sval = va_arg(ap, char *);
			m = strlen(sval);
			break;

		case 'u':
			sprintf(digits, "%u", va_arg(ap, uint));
			m = strlen(digits);
			sval = digits;
			break;

		case 'x':
			sprintf(digits, "%x", va_arg(ap, uint));
			m = strlen(digits);
			sval = digits;
			break;

		case 'd':
			sprintf(digits, "%d", va_arg(ap, int));
			m = strlen(digits);
			sval = digits;
			break;
		}

		memcpy(bufp, sval, m);
		bufp += m;
		len += m;
		continue;
	}

	mysqlcmd[len] = '\0';
	va_end(ap);
	
	if (mysql_real_query(mysqldb, mysqlcmd, len)) {
		if (debug)
			report(LOG_DEBUG, "mysql_db_query: Cannot query. Error - %s", mysql_error(mysqldb));
		free(mysqlcmd);
		return(0);
	}

	free(mysqlcmd);
	return(1);
}

int mysql_db_get_host ( char *field, tac_uint32 nas, char *data, TAC_DBINFO *info )
{
	MYSQL mysqldb;
	MYSQL_RES *res;
	MYSQL_ROW row;
	int ret=1;

	if (debug)
		report(LOG_DEBUG, "mysql_db_get_host: getting %s from nas(%u)", field, nas);

	if (!mysql_db_connect(&mysqldb, info))
		return ret;

	if (!mysql_db_query(&mysqldb, HOSTAUTHSQL, field, nas))
		return ret;

	res = mysql_store_result(&mysqldb);
	if (mysql_num_rows(res) >= 1) {
		while (ret && (row = mysql_fetch_row(res))) {
			if (!tac_empty_string(row[0])) {
				strcpy(data, row[0]);
				if (debug)
					report(LOG_DEBUG, "mysql_db_get_host: got (%s)", data);
				ret = 0;
			}
		}
	}

	mysql_free_result(res);
	mysql_close(&mysqldb);
	return ret;
}

int mysql_db_verify(user, host, clientip, users_passwd, dbinfo, recurse)
char *user, *users_passwd, *clientip;
tac_uint32 host;
TAC_DBINFO *dbinfo;                 
int recurse;
{
    MYSQL mysqldb;
    MYSQL_RES *res;
    MYSQL_ROW row;
    char member[30], password[30];
    int cont=1;

    if (!mysql_db_connect(&mysqldb, dbinfo)) {
	return(0);
    }

    if (host) {
	if (debug & DEBUG_AUTHEN_FLAG)
		report(LOG_DEBUG, "mysql_db_verify: verify host(%u)", host);

	if ((strcmp(dbinfo->field_passwd,"enable")==0) && mysql_db_query(&mysqldb, HOSTAUTHSQL, dbinfo->field_passwd, host)) {
		res = mysql_store_result(&mysqldb);
		if (mysql_num_rows(res)) {
			row = mysql_fetch_row(res);
    			strcpy(password, row[0]);
			mysql_free_result(res);
			if (!tac_empty_string(password))
				goto verify;
		}
		mysql_free_result(res);
	}
    }

    if (debug & DEBUG_AUTHEN_FLAG)
	report(LOG_DEBUG, "mysql_db_verify: verify user(%s)", user);

    if (!mysql_db_query(&mysqldb, AUTHSQL, dbinfo->field_passwd, dbinfo->table, dbinfo->field_name, user)) {
	if (debug & DEBUG_AUTHEN_FLAG)
		report(LOG_DEBUG, "mysql_db_verify: Mysql database problems.");
	mysql_close(&mysqldb);
        return(0);
    }

    res = mysql_store_result(&mysqldb);
    if (mysql_num_rows(res) == 0) {
		if (debug & DEBUG_AUTHEN_FLAG)
			report(LOG_DEBUG, "mysql_db_verify: user(%s) not found");
		mysql_close(&mysqldb);
		return (0);
    }
    row = mysql_fetch_row(res);
    strcpy(password, row[0]);
    strcpy(member, row[1]);
    mysql_free_result(res);

    if (!tac_empty_string(clientip)) {
	if (!mysql_db_query(&mysqldb, USERACCESSSQL, user, member, clientip)) {
	    if (debug & DEBUG_AUTHEN_FLAG)
		report(LOG_DEBUG, "mysql_db_verify: Mysql database problems.");
	    mysql_close(&mysqldb);
            return(0);
    	}
    	res = mysql_store_result(&mysqldb);
    	if (mysql_num_rows(res)) {
	    while (cont && (row = mysql_fetch_row(res))) {
	    	if (strcmp(row[0],"58")==0) {
			    if (debug & DEBUG_AUTHEN_FLAG)
			    	report(LOG_DEBUG, "mysql_db_verify: User is not authorized to access nas(%u) from this address(%s)", host, clientip);
			    mysql_free_result(res);
			    mysql_close(&mysqldb);
			    return(0);
	    	} else if (strcmp(row[0],"57")==0)
			cont=0;
	    }
	    if (debug & DEBUG_AUTHEN_FLAG)
		    report(LOG_DEBUG, "mysql_db_verify: User is authorized to access nas(%u) from this address(%s)", host, clientip);
    	}
    } else {
	    if (debug & DEBUG_AUTHEN_FLAG)
		    report(LOG_DEBUG, "mysql_db_verify: User is login in through console");
    }

    if (tac_empty_string(password)) {
	if (!recurse) {
	    if (debug & DEBUG_AUTHEN_FLAG)
        	report(LOG_DEBUG, "mysql_db_verify: recurse was not requested.");
	    mysql_close(&mysqldb);
            return(0);
        }

	if (tac_empty_string(member)) {
	    if (debug & DEBUG_AUTHEN_FLAG)
        	report(LOG_DEBUG, "mysql_db_verify: User(%s) is not a member of any group.", user);
	        mysql_close(&mysqldb);
		return(0);
	}

	if (debug & DEBUG_AUTHEN_FLAG)
		report(LOG_DEBUG, "mysql_db_verify: building member(%s) query", member);

	if (!mysql_db_query(&mysqldb, AUTHSQL, dbinfo->field_passwd,dbinfo->table, dbinfo->field_name, member)) {
		mysql_close(&mysqldb);
		return(0);
	}
    
	res = mysql_store_result(&mysqldb);
	if (mysql_num_rows(res) == 0) {
		if (debug & DEBUG_AUTHEN_FLAG)
			report(LOG_DEBUG, "mysql_db_verify: member(%s) not found", member);
		mysql_close(&mysqldb);
		return(0);
	}
    	row = mysql_fetch_row(res);
	strcpy(password, row[0]);
	mysql_free_result(res);

	if (tac_empty_string(password)) {
	    if (debug & DEBUG_AUTHEN_FLAG)
		report(LOG_DEBUG, "mysql_db_verify: member password is not found");
	    mysql_close(&mysqldb);
	    return(0);
	}
    }

verify:
    mysql_close(&mysqldb);
  
    if (!des_verify(users_passwd, password)) 
	return (0);
    return (1); 
} 

int 
mysql_db_acct(TAC_DBINFO *dbinfo, char *mysqlcmd)
{

MYSQL mysqldb;
MYSQL_RES *res;
MYSQL_ROW row;
	
   if (debug & DEBUG_ACCT_FLAG)
	report(LOG_DEBUG, "mysql_db_acct: Connecting to server %s and database %s", dbinfo->host, dbinfo->database);

    if (!mysql_db_connect(&mysqldb, dbinfo)) {
	    return(0);
    }


    if (mysql_real_query(&mysqldb,mysqlcmd, strlen(mysqlcmd)))
	{
	if (debug & DEBUG_ACCT_FLAG)
		report(LOG_DEBUG, "mysql_db_acct: cannot query database");
	mysql_close(&mysqldb);
	return(0);
        }

    if ( mysql_affected_rows( &mysqldb ) < 0 )
	{
	if (debug & DEBUG_ACCT_FLAG)
		report(LOG_DEBUG, "mysql_db_acct: Insert isn't sucess");
	mysql_close(&mysqldb);
        return(0);
        }

   if (debug & DEBUG_ACCT_FLAG)
	report(LOG_DEBUG, "mysql_db_acct: Updating mysql was successfull");

    mysql_close(&mysqldb);
    return (1); 
}
*/

#define MYSACCTSQL "INSERT INTO shell_acct (usern,server,client,port," \
				   " rectype,task_id,starttime,timezone," \
		           " elapsed_time,disc_cause,priv_lvl,cmd) " \
				   " VALUES (\"%s\",\"%s\",\"%s\",\"%s\",\"%s\",%s,%s,\"%s\",%s,%s,%s,\"%s\")"

int 
mysql_db_shell_acct(TAC_DBINFO *dbinfo, char *s_name, char *c_name, char *a_username, struct acct_rec *rec)
{
	int sql_len;
	MYSQL mysqldb;
	char mysqlcmd[SQLCMDL];
	char *pCmdFormat = MYSACCTSQL;

	char *pArg_cmd;
	char *pArg_timezone;
	char *pArg_elapsed_time;
	char *pArg_disc_cause;
	char *pArg_priv_lvl;
	char *pArg_port;
	char *pArg_rectype;
	char *pArg_task_id;
	char *pArg_starttime;
	char *pUnknown = "unknown";
	char *pDBNULL = "NULL";
	
	if (debug & DEBUG_ACCT_FLAG)
		report(LOG_DEBUG, "MySQL: write shell accounting record");

/* Connect database server */
	if (!mysql_db_connect(&mysqldb, dbinfo)) {
		return(0);
	}

/* Prepare variables for select string */

	if(NULL == (pArg_cmd = find_attr_value("cmd", rec->args, rec->num_args))) {
		pArg_cmd = "";
	}
	/* edit the cmd field to remove the <cr> sometimes appended to the message */
	if (*pArg_cmd) {
		char *pTmp = (char *)strstr( pArg_cmd, "<cr>");
		if (pTmp) {
			*pTmp = '\0';
		}
	}

	if(NULL == (pArg_timezone = find_attr_value("timezone", rec->args, rec->num_args))) {
		pArg_timezone = "";
	}
	if(NULL == (pArg_priv_lvl = find_attr_value("priv-lvl", rec->args, rec->num_args))) {
		pArg_priv_lvl = pDBNULL;
	}
	if(NULL == (pArg_starttime = find_attr_value("start_time", rec->args, rec->num_args))) {
		pArg_starttime = "0";
	}
	if(NULL == (pArg_task_id = find_attr_value("task_id", rec->args, rec->num_args))) {
		pArg_task_id = "0";
	}
	if(NULL == (pArg_elapsed_time = find_attr_value("elapsed_time", rec->args, rec->num_args))) {
		pArg_elapsed_time = pDBNULL;
	}

	if(NULL == (pArg_disc_cause = find_attr_value("disc-cause-ext", rec->args, rec->num_args))) {
		if(NULL == (pArg_disc_cause = find_attr_value("disc-cause", rec->args, rec->num_args))) {
			pArg_disc_cause = pDBNULL;
		}
	}

	if (NULL == (pArg_port = rec->identity->NAS_port)) {
		pArg_port = pUnknown;
	}

    switch(rec->acct_type) {
		case ACCT_TYPE_UPDATE:
			pArg_rectype = "upd";
			break;
		case ACCT_TYPE_START:
			pArg_rectype = "start";
			break;
		case ACCT_TYPE_STOP:
			pArg_rectype = "stop";
			break;
		default:
			pArg_rectype = pUnknown;
			break;
    }

/* Check buffer overflow for select string */

	sql_len=strlen(dbinfo->table)+strlen(a_username)+
		strlen(s_name)+strlen(c_name)+
		strlen(pArg_port)+strlen(pArg_rectype)+strlen(pArg_starttime)+strlen(pArg_timezone)+
		strlen(pArg_elapsed_time)+strlen(pArg_disc_cause)+strlen(pArg_priv_lvl)+
		strlen(pArg_cmd)+strlen(pArg_task_id)+
		strlen(pCmdFormat);  

	if (debug & DEBUG_ACCT_FLAG)
		report(LOG_DEBUG, "MySQL: Length of Sql command: %d", sql_len);


	if ( sql_len >= SQLCMDL ) {
        if (debug & DEBUG_ACCT_FLAG)
			report(LOG_DEBUG, "MySQL: Sql cmd greater than allowed limit of %d", SQLCMDL);
       	return(0);
	}

	sprintf(mysqlcmd, pCmdFormat, 
		a_username,s_name,c_name,pArg_port, pArg_rectype, pArg_task_id, pArg_starttime, pArg_timezone,
		pArg_elapsed_time, pArg_disc_cause, pArg_priv_lvl, pArg_cmd);

	if (debug & DEBUG_ACCT_FLAG)
		report(LOG_DEBUG, "MySQL: %s ", mysqlcmd);

/*  Query database */

	if (mysql_query(&mysqldb,mysqlcmd)) {
		if (debug & DEBUG_ACCT_FLAG)
			report(LOG_DEBUG, "MySQL: Error %d, %s", mysql_errno(&mysqldb), mysql_error(&mysqldb));
		return(0);
	}

/* Check if accounting is sucess */
    if ( mysql_affected_rows( &mysqldb ) < 0 ) {
		if (debug & DEBUG_ACCT_FLAG)
			report(LOG_DEBUG, "MySQL: Insert isn't sucess");
        return(0);
    }
	return (1); /* Return 1 if verified, 0 otherwise. */
}

/* end of MySQL Shell accounting */

#endif
