# This file created by Andrew Young<baram01@hotmail.com>
# For creating tac_plus database and tables

CREATE DATABASE tacacs;
USE tacacs;

# acl table :-
# id	: ACL identification
# type  : client/host
# seq	: Sequence number
# permission: permit/deny rule in acl
# value	: user/group id or client subnet
# value1 : if client acl, network of the client subnet
#          if host acl, priviledge level
# submask : Subnet mask

CREATE TABLE acl ( id INT(4) NOT NULL, type INT(1) NOT NULL,
	seq INT(4) NOT NULL, permission INT(2) NOT NULL,
	value VARCHAR(20) NOT NULL, value1 REAL, submask REAL,
	PRIMARY KEY (id, type, seq) );
INSERT INTO acl (0, 1, 1, 57, "0.0.0.0/0", 0, 0);
# host table :-
# ip	: IP V4 address of host
# hkey	: Decryption key
# enable: Enable password for host that overrides all other enable passwords.
# prompt: Banner to be displayed on host
# network: Network of the IP
# submask: Subnet mask

CREATE TABLE host (ip varchar(16) NOT NULL PRIMARY KEY, hkey varchar(20),
	enable varchar(35), prompt TEXT, network REAL NOT NULL,
	submask REAL NOT NULL, loginacl INT(4), enableacl INT(4),
	INDEX net (network), INDEX sub (submask) );

# user table :-
# uid        : User/Group id
# gid        : Group id (used if the data is a user otherwise NULL)
# comment    : Description about the user
# password   : Password
# enable     : Enable password
# gpassword  : Global password
# arap       : ARAP
# pap        : PAP
# chap       : CHAP
# mschap     : MSCHAP
# expires    : Expiration date and time
# b_author   : Before authorization
# a_author   : After authorization
# svc_dflt   : Default behaviour for service or command
# maxsess    : Maximum sessions
# user       : 1 - User, 2 - Group
# acl_id     : ACL that limits user/group to specific IP ranges
# sess       : Current number of sessions open

CREATE TABLE user ( uid varchar(20) NOT NULL PRIMARY KEY, gid varchar(20),
	comment text, password varchar(35), enable varchar(35),
	gpassword varchar(35), arap varchar(35), pap varchar(35),
	chap varchar(35), mschap varchar(35), expires datetime,
	b_author varchar(20), a_author varchar(20), svc_dflt int(4),
	maxsess int(4), user int(1), acl_id int(4), sess int(4) );

# contact_info :- ONLY USED FOR THE TRACKING USERS VIA WEB
# uid        : User id
# fname      : First name
# surname    : Surname (last name)
# address1   : Address
# address2   : 
# city       : City
# state      : State
# zip        : Zip code
# phone      : Telephone number
# email	     : Email

CREATE TABLE contact_info ( uid varchar(20) NOT NULL PRIMARY KEY,
	fname varchar(40) NOT NULL, surname varchar(40) NOT NULL,
	address1 varchar(40), address2 varchar(40), city varchar(30),
	state char(2), zip char(5), phone varchar(14), email varchar(100));

# admin table :- ONLY USED FOR WEB
# uid		: User ID
# password	: Password
# priv_lvl	: Priviledge Level
# link		: Link to user table

CREATE TABLE admin ( uid varchar(20) NOT NULL PRIMARY KEY,
	password VARCHAR(35) NOT NULL, priv_lvl INT(2), link INT(1));

INSERT INTO admin VALUES ('admin',ENCRYPT('system'), 15, 0);

# node table :-
# uid        : User ID
# seq        : Sequence number
# service    : Service type (N_svc_cmd, N_svc_exec, N_svc_ppp, etc)
# type       : Type of node (N_arg, N_optarg, N_permit, N_deny, etc)
# value      : value of node
# value1     : value of node

CREATE TABLE node ( uid varchar(20) NOT NULL, seq int(4) NOT NULL,
	service int(4) NOT NULL, type int(4), value varchar(50) NOT NULL,
	value1 varchar(50), INDEX service(uid, service), 
	INDEX command(uid, service, value));

# accounting table :-
# date       : Time stamp of occurance
# nas        : Network Access Server IP(eq. switch)
# uid        : User ID
# terminal   : Terminal used to connect to device
# client_ip  : Client IP
# type       : service type (start, stop, etc..)
# service    : service (exec, shell, etc..)
# priv_lvl   : Priviledge level (usefull in network device)
# cmd        : Command used
# elapsed_time : How much the user spent on router
# bytes_in : Incoming bytes to port
# bytes_out : Outgoing bytes from port
 
CREATE TABLE accounting( date datetime NOT NULL, nas varchar(16) NOT NULL,
	uid varchar(20) NOT NULL, terminal varchar(20),
	client_ip varchar(16) NOT NULL, type varchar(20), service varchar(20),
	priv_lvl INT(2), cmd varchar(255), elapsed_time INT(6),
	bytes_in INT(10), bytes_out INT(10), INDEX date_index(date),
	INDEX acct_index(uid), INDEX nas_index(nas),
	INDEX client_index(client_ip));

# access table :-
# date       : Time stamp of occurance
# nas        : Network Access Server IP(eq. switch)
# terminal   : Terminal used to connect to device
# uid        : User ID
# client_ip  : Client IP
# service    : service (login, enable, etc..)
# status     : rejected/accepted

CREATE TABLE access( date datetime NOT NULL, nas VARCHAR(16) NOT NULL,
	terminal VARCHAR(20), uid VARCHAR(20) NOT NULL,
	client_ip VARCHAR(16) NOT NULL, service VARCHAR(10), status VARCHAR(10),
	INDEX date_index(date), INDEX nas_index(nas), INDEX uid_index(uid),
	INDEX client_index(client_ip) );
 
#create users needed to administrate tacacs

GRANT ALL ON tacacs.* TO tacacs@localhost IDENTIFIED BY 'tac_plus';
GRANT ALL ON tacacs.* TO tacacs IDENTIFIED BY 'tac_plus';

